#include <cmath>
#include <limits>
#include "nr.h"

//___________________________________________________________
// вычисление хи-квадрат распределения ХИ2РАСП(x; степени_свободы)
// on error (wrong arguments) return -1;
DP NR::chi2distr(DP xi2, DP df){
	DP pval = gammq(df*0.5, xi2*0.5);
	return pval;
}
//___________________________________________________________
// 6.2 Incomplete gamma function
// BOOK: Numerical recipes in C. 2-nd edition
// Q(a,x) = gammq(a,x)=1/Г(а)*Int(x,inf)(exp(-t)*t^(a-1)dt)
// 0 <= Q <= 1
// Qchi(xi2, df) = Q(df/2, xi2/2)   -> chi-square distribution
// Qchi(xi2=0, df)=1 ,  Qchi(xi2=inf, df)=0
// on error (wrong arguments) return -1;
DP NR::gammq(const DP a, const DP x)
{
	DP gamser,gammcf,gln;

	if (x < 0.0 || a <= 0.0){
		nrerror("Invalid arguments in routine gammq");
		return -1;
	}
	if (x < a+1.0) {
		gser(gamser,a,x,gln);
		return 1.0-gamser;
	} else {
		gcf(gammcf,a,x,gln);
		return gammcf;
	}
}
//_______________________________________
void NR::gser(DP &gamser, const DP a, const DP x, DP &gln)
{
	const int ITMAX=100;
	const DP EPS=numeric_limits<DP>::epsilon();
	int n;
	DP sum,del,ap;

	gln=gammln(a);
	if (x <= 0.0) {
		if (x < 0.0) nrerror("x less than 0 in routine gser");
		gamser=0.0;
		return;
	} else {
		ap=a;
		del=sum=1.0/a;
		for (n=0;n<ITMAX;n++) {
			++ap;
			del *= x/ap;
			sum += del;
			if (fabs(del) < fabs(sum)*EPS) {
				gamser=sum*exp(-x+a*log(x)-gln);
				return;
			}
		}
		nrerror("a too large, ITMAX too small in routine gser");
		return;
	}
}
//_______________________________________
void NR::gcf(DP &gammcf, const DP a, const DP x, DP &gln)
{
	const int ITMAX=100;
	const DP EPS=numeric_limits<DP>::epsilon();
	const DP FPMIN=numeric_limits<DP>::min()/EPS;
	int i;
	DP an,b,c,d,del,h;

	gln=gammln(a);
	b=x+1.0-a;
	c=1.0/FPMIN;
	d=1.0/b;
	h=d;
	for (i=1;i<=ITMAX;i++) {
		an = -i*(i-a);
		b += 2.0;
		d=an*d+b;
		if (fabs(d) < FPMIN) d=FPMIN;
		c=b+an/c;
		if (fabs(c) < FPMIN) c=FPMIN;
		d=1.0/d;
		del=d*c;
		h *= del;
		if (fabs(del-1.0) <= EPS) break;
	}
	if (i > ITMAX) nrerror("a too large, ITMAX too small in gcf");
	gammcf=exp(-x+a*log(x)-gln)*h;
}
//_______________________________________
DP NR::gammln(const DP xx)
{
	int j;
	DP x,y,tmp,ser;
	static const DP cof[6]={76.18009172947146,-86.50532032941677,
		24.01409824083091,-1.231739572450155,0.1208650973866179e-2,
		-0.5395239384953e-5};

	y=x=xx;
	tmp=x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser=1.000000000190015;
	for (j=0;j<6;j++) ser += cof[j]/++y;
	return -tmp+log(2.5066282746310005*ser/x);
}
